import { defineQuery, hasComponent } from 'bitecs'
import { query as q } from './query-parser/index.js'
import { parseAll } from './parser.js'
import { curry, of, is, reduce, flip, concat, map, when, assoc, mergeLeft, isNil, path, prop, ifElse, diverge, pipe, curryN, always, tap } from './fn.js'
import { Relationship } from './query-parser/types.js'

/*
 * necessary engine structure
 * worlds: World[]
 * Component: Component[]
 */

const prepareQuery = (query, component) => {
  const { match, returnValues } = query
  const relationship = is(Relationship, match)

  if (relationship) {
    const { from: f, to: t, edge: e } = match
    const [from, to, edge] = map(q => prepareQuery({ match: q, returnValues }, component), [f, t, e])

    return { from, to, edge, relationship, schema: query }
  } else {
    const label = match.label?.value?.value
    const name = match.name?.value?.value
    const type = component[label]

    return { name, label, type, relationship, query: defineQuery([type]), schema: query }
  }
}

const getEndNode = end => end.relationship ? end.from : end

const matches = (world, type, query, set = new Set()) => entity => {
  if (set.has(entity)) {
    return true
  } else if (hasComponent(world, type, entity) && query.includes(entity)) {
    set.add(entity)
    return true
  } else {
    return false
  }
}

const queryRelationship = (from, edge, to, world) => {
  const matchesFrom = matches(world, from.type, from.results)
  const matchesTo = matches(world, to.type, to.results)

  const Edge = edge.type

  const result = []
  for (let i = 0; i < edge.results.length; i++) {
    const eid = edge.results[i]

    if (matchesFrom(Edge.from[eid]) && matchesTo(Edge.to[eid])) {
      result.push(eid)
    }
  }

  return result
}

const cat = flip(concat)
const assembleQuery = (query, entities) => map(entity => of({ entity, query }), entities)

const executeQuery = curry((query, world) => {
  if (!query.relationship) {
    return assembleQuery(query, query.query(world))
  }

  const { from, to: _to, edge } = query
  const to = getEndNode(_to)

  const Edge = edge.type

  const edges = queryRelationship(
    { type: from.type, results: from.query(world) },
    { type: Edge, results: edge.query(world) },
    { type: to.type, results: to.query(world) },
    world
  )

  if (_to.relationship) {
    return reduce((acc, eid) => {
      const next = {
        ..._to,
        from: {
          ...to,
          query: always(of(Edge.to[eid]))
        }
      }

      return pipe(
        executeQuery(next),
        map(cat([eid])),
        map(([l, r]) => ([{ entity: l, query }, r])),
        x => ([...acc, ...x]),
      )(world)
    }, [], edges)
  } else {
    return assembleQuery(query, edges)
  }
})

const returnValues = pipe(path('query.schema.returnValues.values'), map(prop('value')))
const includes = curry((val, arr) => !isNil(val) && arr.includes(val))

const maybeAssoc = curry((pred, key, value, obj) =>
  when(pred, assoc(key, value), obj)
)

const resolveNode = curry((returns, name, entity, obj) =>
  maybeAssoc(always(includes(name, returns)), name, entity, obj)
)

const resolveRelationship = edges => {
  const { entity, query } = edges
  const Edge = query.edge.type
  const returns = returnValues(edges)

  const queryProp = flip(prop)(query)
  const [from, edge, to] = map(queryProp, ['from', 'edge', 'to'])
  const [fn, en] = [from.name, edge.name]
  const tn = ifElse(prop('relationship'),
    path('from.name'),
    prop('name'),
    to
  )

  return pipe(
    resolveNode(returns, fn, Edge.from[entity]),
    resolveNode(returns, en, entity),
    resolveNode(returns, tn, Edge.to[entity]),
  )({})
}

const resolveReturns = map(
  pipe(
    map(ifElse(
      path('query.relationship'),
      resolveRelationship,
      diverge(resolveNode, [returnValues, path('query.schema.match.name.value.value'), prop('entity'), always({})])
    )),
    reduce(mergeLeft, {})
  )
)

export const query = curry((input, engine) => {
  return parseAll(q, input).map(({ use, ...rest }) => {
    const worldName = use && use.value || 'default'
    const world = engine.world[worldName]
    const prepared = prepareQuery(rest, engine.component)
    const edges = executeQuery(prepared, world)
    return resolveReturns(edges)
  })
})

